// Inf2C Computer Systems, Processor design practical
// Copyright 2004-2013, School of Informatics, The University of Edinburgh

// Control unit for processor. 

// Template file. 
// Only contains implementation for j, addi,lw and halt instructions.

#include "systemc.h"
#include "controlUnit.h"
#include "defines.h"

#define MAX_CYCLES 200

// Control unit for processor. 
void controlUnit::ctrl_regs()
{
    sc_uint<6> t_next_state;

    while (1) {
        wait(); // wait for reset or clock

        if (reset) {
            wait(1, SC_NS);
            cur_state = 0;
            cycle_count = 0;
        } else {
            if (halt.read()) {
                // endSimulation
                std::cout << sc_time_stamp()
                          <<" Halting simulation."
                          << std::endl
                          << "Cycles = "
                          << cycle_count
                          << std::endl;
                sc_stop();
                // No return instruction needed.
            } else if (cycle_count >= MAX_CYCLES) {
                std::cout << sc_time_stamp()
                          <<" Runaway program. Halting simulation!"
                          << std::endl
                          << "Cycles = "
                          << cycle_count
                          << std::endl;
                sc_stop();
            }
            cycle_count++;
            t_next_state = next_state;
            wait(1, SC_NS);
            cur_state = t_next_state;
        }
    }
}



void controlUnit::ctrl_comb()
{
    sc_int<32> t_ir = ir;

    // Extract useful subfields of the instruction
    opcode = t_ir.range(31,26);
    subfunct = t_ir.range(5,0);

    // Default values for various control fields

    halt = false;
    mem_rd = false;
    mem_wrt = false;

    ldPC = false; 
    ldIR = false; 
    ldReg = false;

    ldMAR = false;
    ldALUout = false;
    ldMDLR = false;
    // ldMDSR = false;
    ldA = false;
    ldB = false;

    // Defaults for multiplexer settings and ALU function.
    // While defaults are set here, the main code should always set these
    // explicitly to make intent clear.

    rreg_addr_mux = RA_A;  
    wreg_addr_mux = WA_RD;  
    pc_mux = PC_ALU; 
    addr_mux = ADDR_PC;  
    a_mux = A_PC;
    b_mux = B_REG;
    reg_write_mux = RW_ALU_OUT; 
    func = ADD;

    switch(cur_state.read()){

    //----------------------------------------------------------------------
    // STATE 0
    //----------------------------------------------------------------------
    // Instruction fetch.  (All instructions)
    case(0): 
        // Set outputs
        mem_rd = true;
        ldPC = true;
        ldIR = true;
        pc_mux=PC_ALU;
        addr_mux = ADDR_PC;
        a_mux = A_PC;
        b_mux = B_4;
        func = ADD;

        // Set next state
        next_state = 1; 
        break;
    //----------------------------------------------------------------------
    // STATE 1
    //----------------------------------------------------------------------
    // Load A reg (All instructions)
    case(1): 
        // Set outputs
        ldA = true;
        rreg_addr_mux = RA_A;

        // Set next state        
        if (opcode == 0x0 && subfunct == 0xc) {
            // halt instruction
            next_state = 2;  
        }
        else if (opcode == 0x2) {
            // jump instruction
            next_state = 3;  
        }
        else if (opcode == 0x8) {
            // addi instruction
            next_state = 4;
        }
        else if (opcode == 0x23) {
            // lw 
            next_state = 6;
        }
        else {
            // unrecognised instruction
            std::cout << "At " << sc_time_stamp() << ": ";
            printf("unrecognized instruction (opcode=0x%x, subfunct=0x%x)",
                   opcode,
                   subfunct);
            std::cout << " when in state 1" << std::endl;
            halt = true;
        }
        break;
    //----------------------------------------------------------------------
    // STATE 2
    //----------------------------------------------------------------------
    // Halt (halt instruction)
    case (2):
        printf("Halt instruction executed\n");
        halt = true;
        break;
    //----------------------------------------------------------------------
    // STATE 3
    //----------------------------------------------------------------------
    // Set PC to jump address (j instruction)
    case (3):
        // Set outputs
        ldPC = true;
        pc_mux = PC_IMM;

        // Set next state
        next_state = 6;
        break;

    //----------------------------------------------------------------------
    // STATE 4
    //----------------------------------------------------------------------
    // Set ALUout to addition result A_reg+sign-extend(IR[15-0])  (addi)
    case (4):
        // Set outputs
        ldALUout = true;
        a_mux = A_REG;
        b_mux = B_IR_16;
        func = ADD;
        // Set next state
        next_state = 5;
        break;

    //----------------------------------------------------------------------
    // STATE 5
    //----------------------------------------------------------------------
    // Store ALUout to Register File (addi instruction)
    case (5):
        // Set outputs
        ldReg = true;
        wreg_addr_mux = WA_RT;
        reg_write_mux = RW_ALU_OUT;

        // Set next state
        next_state = 0;
        break;

    //----------------------------------------------------------------------
    // STATE 6
    //----------------------------------------------------------------------
    // Set MAR to computed address (lw instruction)
    case (6):
        // Set outputs
        ldMAR = true;
        a_mux = A_REG;
        b_mux = B_IR_16;
        func = ADD;

        // Set next state
        next_state = 7;
        break;


    //----------------------------------------------------------------------
    // STATE 7
    //----------------------------------------------------------------------
    // Fetch data word from memory (lw instruction)
    case (7):
        // Set outputs
        mem_rd = true;
        ldMDLR = true;
        addr_mux = ADDR_MAR;

        // Set next state
        next_state = 8;
        break;


    //----------------------------------------------------------------------
    // STATE 8
    //----------------------------------------------------------------------
    // Write data to Register File (lw instruction)
    case (8):
        // Set outputs
        ldReg = true;
        wreg_addr_mux = WA_RT;
        reg_write_mux = RW_MEM;

        // Set next state
        next_state = 0;
        break;


    //----------------------------------------------------------------------
    // OTHERWISE
    //----------------------------------------------------------------------
    default: 
        std::cout << "At " << sc_time_stamp() << ": ";
        std::cout << "unexpected current state " << cur_state.read();
        std::cout << std::endl;
        halt = true;

    } // end switch(cur_state)

    return;
}
